# This Build.PL for OpenTelemetry-Instrumentation-namespace was generated by Dist::Zilla::Plugin::ModuleBuildTiny 0.020.
use strict;
use warnings;


use v5.38.0;
use Module::Build::Tiny 0.034;
Build_PL();

